
function inRange(aNumber, aLow, aHigh)
  return (aNumber >= aLow) and (aNumber <= aHigh)
end

function strToSquare(aStr)
  assert(#aStr == 2)
  assert(string.match(aStr, '[a-h][1-8]'))
  
  return
    string.byte(aStr, 1) - string.byte('a') + 1,
    string.byte(aStr, 2) - string.byte('1') + 1
end

function squareToStr(aX, aY)
  assert(inRange(aX, 1, 8))
  assert(inRange(aY, 1, 8))
  
  return string.char(
    string.byte('a') + aX - 1,
    string.byte('1') + aY - 1
  )
end

function strToMove(aStr)
  assert(inRange(#aStr, 4, 5))
  assert(string.match(aStr, '[a-h][1-8][a-h][1-8]' .. ((#aStr == 5) and '[nbrq]' or '')))
  
  local promotion = (#aStr == 5) and string.sub(aStr, 5, 5) or nil
  
  return
    string.byte(aStr, 1) - string.byte('a') + 1,
    string.byte(aStr, 2) - string.byte('1') + 1,
    string.byte(aStr, 3) - string.byte('a') + 1,
    string.byte(aStr, 4) - string.byte('1') + 1,
    promotion
end

function moveToStr(aX1, aY1, aX2, aY2)
  assert(inRange(aX1, 1, 8))
  assert(inRange(aY1, 1, 8))
  assert(inRange(aX2, 1, 8))
  assert(inRange(aY2, 1, 8))
  
  return string.char(
    string.byte('a') + aX1 - 1,
    string.byte('1') + aY1 - 1,
    string.byte('a') + aX2 - 1,
    string.byte('1') + aY2 - 1
  )
end

function boardToText(aBoard)
  local result = ''
  for y = 8, 1, -1 do
    for x = 1, 8 do
      result = result .. ((aBoard[x][y] ~= nil) and aBoard[x][y] or '.')
    end
    if y > 1 then
      result = result .. '\n'
    end
  end
  return result
end

function movePiece(aBoard, x1, y1, x2, y2, aPromotion)
  assert(inRange(x1, 1, 8))
  assert(inRange(y1, 1, 8))
  assert(inRange(x2, 1, 8))
  assert(inRange(y2, 1, 8))
  
  if aBoard[x1][y1] == nil then
    return false
  else
    aBoard[x2][y2] = aPromotion or aBoard[x1][y1]
    aBoard[x1][y1] = nil
    return true 
  end
end

function isColor(aBoardValue, aColor)
  assert(string.match(aColor, '[wb]'))
  
  return (aBoardValue ~= nil) and string.match(aBoardValue, (aColor == 'w') and '[PNBRQK]' or '[pnbrqk]')
end

function otherColor(aColor)
  assert(string.match(aColor, '[wb]'))
  
  return (aColor == 'w') and 'b' or 'w'
end

function isWhitePiece(aBoardValue)
  return isColor(aBoardValue, 'w')
end

function isPawn(aBoardValue)
  return (aBoardValue == 'P') or (aBoardValue == 'p')
end

function isKnight(aBoardValue)
  return (aBoardValue == 'N') or (aBoardValue == 'n')
end

function isBishop(aBoardValue)
  return (aBoardValue == 'B') or (aBoardValue == 'b')
end

function isRook(aBoardValue)
  return (aBoardValue == 'R') or (aBoardValue == 'r')
end

function isQueen(aBoardValue)
  return (aBoardValue == 'Q') or (aBoardValue == 'q')
end

function isKing(aBoardValue)
  return (aBoardValue == 'K') or (aBoardValue == 'k')
end

function strToBoard(aPiecePlacementStr)
  local result = {{}, {}, {}, {}, {}, {}, {}, {}}
  local i, x, y = 1, 1, 8
  while i <= #aPiecePlacementStr do
    local s = string.sub(aPiecePlacementStr, i, i)
    if s == '/' then
      y = y - 1
      x = 1
    elseif string.match(s, '%d') then
      for i = 1, tonumber(s) do
        result[x][y] = nil
        x = x + 1
      end
    else
      result[x][y] = s
      x = x + 1
    end
    i = i + 1
  end
  return result
end

function encodePositionData(aFENRecord)
  local t = {}
  for s in string.gmatch(aFENRecord or 'rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1', '%S+') do
    t[#t + 1] = s
  end
  assert(#t == 6)
  return {
    piecePlacement = strToBoard(t[1]),
    activeColor = t[2],
    castlingAvailability = t[3],
    enPassantTargetSquare = t[4],
    halfmoveClock = tonumber(t[5]),
    fullmoveNumber = tonumber(t[6])
  }
end

function boardToStr(aBoard)
  local result = ''
  for y = 8, 1, -1 do
    local x = 1
    while x <= 8 do
      if aBoard[x][y] ~= nil then
        result = result .. aBoard[x][y]
        x = x + 1
      else
        local n = 0
        while (x <= 8) and (aBoard[x][y] == nil) do
          n = n + 1
          x = x + 1
        end
        result = result .. tostring(n)
      end
    end
    if y > 1 then
      result = result .. '/'
    end
  end
  return result
end

function decodePositionData(aPositionData)
  return string.format(
    '%s %s %s %s %d %d',
    boardToStr(aPositionData.piecePlacement),
    aPositionData.activeColor,
    aPositionData.castlingAvailability,
    aPositionData.enPassantTargetSquare,
    aPositionData.halfmoveClock,
    aPositionData.fullmoveNumber
  )
end

local vectors = {
  {x =-1, y = 1},
  {x = 1, y = 1},
  {x =-1, y =-1},
  {x = 1, y =-1},
  {x =-1, y = 0},
  {x = 1, y = 0},
  {x = 0, y = 1},
  {x = 0, y =-1},
  {x = 1, y = 2},
  {x = 2, y = 1},
  {x = 2, y =-1},
  {x = 1, y =-2},
  {x =-1, y =-2},
  {x =-2, y =-1},
  {x =-2, y = 1},
  {x =-1, y = 2}
}

function computeTargetSquare(aX1, aY1, aVectorIndex)
  assert(inRange(aVectorIndex, 1, #vectors))
  
  local x2, y2 =
    aX1 + vectors[aVectorIndex].x,
    aY1 + vectors[aVectorIndex].y
  
  if inRange(x2, 1, 8) and inRange(y2, 1, 8) then
    return true, x2, y2
  else
    return false
  end
end

function generateMoves1(aBoard, aColor)
  local index1, index2
  local result = {}
  for x = 1, 8 do
    for y = 1, 8 do
      if isColor(aBoard[x][y], aColor) then
        if isPawn(aBoard[x][y]) then
          if isWhitePiece(aBoard[x][y]) then
            index1, index2 = 1, 2
          else
            index1, index2 = 3, 4
          end
          for vectorIndex = index1, index2 do
            local success, x2, y2 = computeTargetSquare(x, y, vectorIndex)
            if success and isColor(aBoard[x2][y2], otherColor(aColor)) then
              result[#result + 1] = moveToStr(x, y, x2, y2)
            end
          end
        elseif isKnight(aBoard[x][y]) or isKing(aBoard[x][y]) then
          if isKnight(aBoard[x][y]) then
            index1, index2 = 9, 16
          elseif isKing(aBoard[x][y]) then
            index1, index2 = 1, 8
          end
          for vectorIndex = index1, index2 do
            local success, x2, y2 = computeTargetSquare(x, y, vectorIndex)
            if success and not isColor(aBoard[x2][y2], aColor) then
              result[#result + 1] = moveToStr(x, y, x2, y2)
            end
          end
        elseif isBishop(aBoard[x][y]) or isRook(aBoard[x][y]) or isQueen(aBoard[x][y]) then
          if isBishop(aBoard[x][y]) then
            index1, index2 = 1, 4
          elseif isRook(aBoard[x][y]) then
            index1, index2 = 5, 8
          elseif isQueen(aBoard[x][y]) then
            index1, index2 = 1, 8
          end
          for vectorIndex = index1, index2 do
            local success, x2, y2 = computeTargetSquare(x, y, vectorIndex)
            while success and not isColor(aBoard[x2][y2], aColor) do
              result[#result + 1] = moveToStr(x, y, x2, y2)
              if aBoard[x2][y2] ~= nil then
                break
              end
              success, x2, y2 = computeTargetSquare(x2, y2, vectorIndex)
            end
          end
        end
      end
    end
  end
  return result
end

function think(aPositionData)
  local moves = generateMoves1(aPositionData.piecePlacement, otherColor(aPositionData.activeColor))
  local check = false
  local castlingCheck = {
    e1g1 = false,
    e1c1 = false,
    e8g8 = false,
    e8c8 = false
  }
  
  for k, v in ipairs(moves) do
    local x2, y2 = strToSquare(string.sub(v, 3, 4))
    
    check = check or isKing(aPositionData.piecePlacement[x2][y2])
    
    local row = (aPositionData.activeColor == 'w') and 1 or 8
    
    if y2 == row then
      if inRange(x2, 5, 8) then
        if row == 1 then
          castlingCheck.e1g1 = true
        else
          castlingCheck.e8g8 = true
        end
      else
        if row == 1 then
          castlingCheck.e1c1 = true
        else
          castlingCheck.e8c8 = true
        end
      end
    end
  end
  
  local result = {
    check = check and true or false,
    castlingCheck = castlingCheck
  }
  return result
end

function isWayClear(aBoard, aColor, aKingX, aRookX)
  local result = true
  local y = (aColor == 'w') and 1 or 8
  
  result = result and (aBoard[aKingX][y] == (aColor == 'w') and 'K' or 'k')
  result = result and (aBoard[aRookX][y] == (aColor == 'w') and 'R' or 'r')
  
  local index1 = (aRookX == 8) and (aKingX + 1) or (aKingX - 1)
  local index2 = (aRookX == 8) and (aRookX - 1) or (aRookX + 1)
  local step = (index2 > index1) and 1 or -1
  
  for x = index1, index2, step do
    result = result and (aBoard[x][y] == nil)
  end
  
  return result
end

function generateMoves2(aPositionData, aColor)
  local index1, index2
  local result = {}
  
  local extraPositionData = think(aPositionData)
  
  local function generateCastling(aSymbol, aColor, aKingX, aRookX, aMove)
    if string.match(aPositionData.castlingAvailability, aSymbol)
    and isWayClear(aPositionData.piecePlacement, aColor, aKingX, aRookX)
    and not extraPositionData.castlingCheck[aMove] then
      result[#result + 1] = aMove
    end
  end
  
  for x = 1, 8 do
    for y = 1, 8 do
      if isColor(aPositionData.piecePlacement[x][y], aColor) then
        
        if isPawn(aPositionData.piecePlacement[x][y]) then
          if isWhitePiece(aPositionData.piecePlacement[x][y]) then
            index1 = 7
          else
            index1 = 8
          end
          
          local success, x2, y2 = computeTargetSquare(x, y, index1)
          if success and (aPositionData.piecePlacement[x2][y2] == nil) then
            result[#result + 1] = moveToStr(x, y, x2, y2)
            if y == ((aColor == 'w') and 2 or 7) then
              success, x2, y2 = computeTargetSquare(x2, y2, index1)
              if success and (aPositionData.piecePlacement[x2][y2] == nil) then
                result[#result + 1] = moveToStr(x, y, x2, y2)
              end
            end
          end
          
          if isWhitePiece(aPositionData.piecePlacement[x][y]) then
            index1, index2 = 1, 2
          else
            index1, index2 = 3, 4
          end
          for vectorIndex = index1, index2 do
            local success, x2, y2 = computeTargetSquare(x, y, vectorIndex)
            if success
            and (aPositionData.piecePlacement[x2][y2] == nil)
            and (squareToStr(x2, y2) == aPositionData.enPassantTargetSquare) then
              result[#result + 1] = moveToStr(x, y, x2, y2)
            end
          end
        
        elseif isKing(aPositionData.piecePlacement[x][y]) and not extraPositionData.check then
          if isWhitePiece(aPositionData.piecePlacement[x][y]) then
            generateCastling('K', aColor, 5, 8, 'e1g1')
            generateCastling('Q', aColor, 5, 1, 'e1c1')
          else
            generateCastling('k', aColor, 5, 8, 'e8g8')
            generateCastling('q', aColor, 5, 1, 'e8c8')
          end
        end
      end
    end
  end
  return result
end

function removeCastling(aPositionData, aChar)
  local result = string.gsub(aPositionData.castlingAvailability, aChar, '')
  if result == '' then
    result = '-'
  end
  aPositionData.castlingAvailability = result
end

function doMove(aPositionData, x1, y1, x2, y2, aPromotion)
  assert(aPositionData.piecePlacement[x1][y1] ~= nil)
  
  local result = true
  
  if isKing(aPositionData.piecePlacement[x1][y1]) and (x1 == 5) then
    if (y1 == 1) and (aPositionData.activeColor == 'w') then
      removeCastling(aPositionData, 'K')
      removeCastling(aPositionData, 'Q')
    elseif (y1 == 8)  and (aPositionData.activeColor == 'b') then
      removeCastling(aPositionData, 'k')
      removeCastling(aPositionData, 'q')
    end
  end
  
  if isRook(aPositionData.piecePlacement[x1][y1]) then
    if (y1 == 1) and (aPositionData.activeColor == 'w') then
      if (x1 == 8) then
        removeCastling(aPositionData, 'K')
      end
      if (x1 == 1) then
        removeCastling(aPositionData, 'Q')
      end
    elseif (y1 == 8) and (aPositionData.activeColor == 'b') then
      if (x1 == 8) then
        removeCastling(aPositionData, 'k')
      end
      if (x1 == 1) then
        removeCastling(aPositionData, 'q')
      end
    end
  end
  
  if isPawn(aPositionData.piecePlacement[x1][y1]) and (math.abs(y2 - y1) == 2) then
    aPositionData.enPassantTargetSquare = squareToStr(x1, (aPositionData.activeColor == 'w') and 3 or 6)
  else
    aPositionData.enPassantTargetSquare = '-'
  end
  
  if isKing(aPositionData.piecePlacement[x1][y1]) and (math.abs(x2 - x1) == 2) then
    if x2 == 7 then
      result = result and movePiece(aPositionData.piecePlacement, 8, y1, 6, y2)
    elseif x2 == 3 then
      result = result and movePiece(aPositionData.piecePlacement, 1, y1, 4, y2)
    else
      assert(false)
    end
  end
  
  if isPawn(aPositionData.piecePlacement[x1][y1]) and (math.abs(x2 - x1) == 1) and (aPositionData.piecePlacement[x2][y2] == nil) then
    aPositionData.piecePlacement[x2][y1] = nil
  end
  
  if isPawn(aPositionData.piecePlacement[x1][y1]) or (aPositionData.piecePlacement[x2][y2] ~= nil) then
    aPositionData.halfmoveClock = 0
  else
    aPositionData.halfmoveClock = aPositionData.halfmoveClock + 1
  end
  
  if aPositionData.activeColor == 'b' then
    aPositionData.fullmoveNumber = aPositionData.fullmoveNumber + 1
  end
  
  if isPawn(aPositionData.piecePlacement[x1][y1]) and ((y2 == 1) or (y2 == 8)) then
    if aPromotion == nil then
      aPromotion = (aPositionData.activeColor == 'w') and 'Q' or 'q'
    else
      aPromotion = (aPositionData.activeColor == 'w') and string.upper(aPromotion) or string.lower(aPromotion)
    end
  else
    aPromotion = nil
  end
  
  result = result and movePiece(aPositionData.piecePlacement, x1, y1, x2, y2, aPromotion)
  
  aPositionData.activeColor = otherColor(aPositionData.activeColor)
  
  return result
end

function generateLegalMoves(aPositionData)
  local table1 = generateMoves1(aPositionData.piecePlacement, aPositionData.activeColor)
  local table2 = generateMoves2(aPositionData, aPositionData.activeColor)
  local table3 = {}
  for k, v in ipairs(table1) do table3[#table3 + 1] = v end
  for k, v in ipairs(table2) do table3[#table3 + 1] = v end
  local result = {}
  for k, v in ipairs(table3) do
    local x1, y1, x2, y2 = strToMove(v)
    local afterMovePosition = encodePositionData(decodePositionData(aPositionData))
    if doMove(afterMovePosition, x1, y1, x2, y2, nil) then
      afterMovePosition.activeColor = otherColor(afterMovePosition.activeColor)
      local extraPositionData = think(afterMovePosition)
      if not extraPositionData.check then
        result[#result + 1] = v
      end
    end
  end
  table.sort(result)
  return result
end

function isCastling(aPositionData, aMove)
  local x1, y1, x2, y2 = strToMove(aMove)
  if isKing(aPositionData.piecePlacement[x1][y1])
  and (math.abs(x2 - x1) == 2) then
    local rookMove = string.format('%s%d%s%d', (x2 == 7) and 'h' or 'a', y1, (x2 == 7) and 'f' or 'd', y1)
    return true, rookMove
  else
    return false
  end
end

function isEnPassant(aPositionData, aMove)
  local x1, y1, x2, y2 = strToMove(aMove)
  if isPawn(aPositionData.piecePlacement[x1][y1])
  and (x2 ~= x1)
  and (aPositionData.piecePlacement[x2][y2] == nil) then
    return true, squareToStr(x2, y1)
  else
    return false
  end
end

function isPromotion(aPositionData, aMove)
  local x1, y1, x2, y2 = strToMove(aMove)
  return isPawn(aPositionData.piecePlacement[x1][y1]) and ((y2 == 1) or (y2 == 8))
end
