
require('chessgame')
require('language')

function erase(aTable)
  for k, v in pairs(aTable) do
    aTable[k] = nil
  end
end

local positionData = encodePositionData()
local legalMoves = generateLegalMoves(positionData)

function isLegalMove(aMove)
  for k, v in ipairs(legalMoves) do
    if v == aMove then
      return true
    end
  end
  return false
end

local images = {}
local pieces = {
  {x = 0 * 40, y = 0 * 40, id = 'br'},
  {x = 1 * 40, y = 0 * 40, id = 'bn'},
  {x = 2 * 40, y = 0 * 40, id = 'bb'},
  {x = 3 * 40, y = 0 * 40, id = 'bq'},
  {x = 4 * 40, y = 0 * 40, id = 'bk'},
  {x = 5 * 40, y = 0 * 40, id = 'bb'},
  {x = 6 * 40, y = 0 * 40, id = 'bn'},
  {x = 7 * 40, y = 0 * 40, id = 'br'},
  {x = 0 * 40, y = 1 * 40, id = 'bp'},
  {x = 1 * 40, y = 1 * 40, id = 'bp'},
  {x = 2 * 40, y = 1 * 40, id = 'bp'},
  {x = 3 * 40, y = 1 * 40, id = 'bp'},
  {x = 4 * 40, y = 1 * 40, id = 'bp'},
  {x = 5 * 40, y = 1 * 40, id = 'bp'},
  {x = 6 * 40, y = 1 * 40, id = 'bp'},
  {x = 7 * 40, y = 1 * 40, id = 'bp'},
  {x = 0 * 40, y = 6 * 40, id = 'wp'},
  {x = 1 * 40, y = 6 * 40, id = 'wp'},
  {x = 2 * 40, y = 6 * 40, id = 'wp'},
  {x = 3 * 40, y = 6 * 40, id = 'wp'},
  {x = 4 * 40, y = 6 * 40, id = 'wp'},
  {x = 5 * 40, y = 6 * 40, id = 'wp'},
  {x = 6 * 40, y = 6 * 40, id = 'wp'},
  {x = 7 * 40, y = 6 * 40, id = 'wp'},
  {x = 0 * 40, y = 7 * 40, id = 'wr'},
  {x = 1 * 40, y = 7 * 40, id = 'wn'},
  {x = 2 * 40, y = 7 * 40, id = 'wb'},
  {x = 3 * 40, y = 7 * 40, id = 'wq'},
  {x = 4 * 40, y = 7 * 40, id = 'wk'},
  {x = 5 * 40, y = 7 * 40, id = 'wb'},
  {x = 6 * 40, y = 7 * 40, id = 'wn'},
  {x = 7 * 40, y = 7 * 40, id = 'wr'}
}
local draggable = 0
local move = ''
local offset = {x = 0, y = 0}
local text = literals.ltWhiteToMove
local gameOver = false
local cursorGrab, cursorGrabbing
local validTargetSquares = {}

function imagePath(aImageName)
  return string.format('pictures/%s.png', aImageName)
end

function love.load()
  images['board'] = love.graphics.newImage(imagePath('board'))
  local colors, types = 'wb', 'pnbrqk'
  for i = 1, #colors do
    for j = 1, #types do
      local name = string.sub(colors, i, i) .. string.sub(types, j, j)
      images[name] = love.graphics.newImage(imagePath(name))
    end
  end
  images['valid'] = love.graphics.newImage(imagePath('valid'))
  love.graphics.setBackgroundColor(255, 255, 255)
  cursorGrab = love.mouse.newCursor('pictures/grab.png', 0, 0)
  cursorGrabbing = love.mouse.newCursor('pictures/grabbing.png', 0, 0)
  love.mouse.setCursor(cursorGrab)
end

function screenToBoard(aX, aY)
  return math.floor(aX / 40) + 1, 8 - math.floor(aY / 40)
end

function boardToScreen(aX, aY)
  return 40 * (aX - 1), 40 * (8 - aY)
end

function align(aCoord)
  return math.floor(aCoord / 40 + 0.5) * 40
end

function afterMoveProcedure()
  if positionData.halfmoveClock == 50 then
    text = literals.ltDraw
    gameOver = true
  else
    legalMoves = generateLegalMoves(positionData)
    print(table.concat(legalMoves, '\n') .. '\n')
    local extraPositionData = think(positionData)
    if #legalMoves == 0 then
      text = extraPositionData.check and literals.ltCheckmate or literals.ltStalemate
      if extraPositionData.check then
        text = string.format(
          '%s%s',
          text,
          positionData.activeColor == 'w' and literals.ltBlackWins or literals.ltWhiteWins
        )
      end
      gameOver = true
    else
      text = string.format(
        '%s%s',
        extraPositionData.check and literals.ltCheck or '',
        positionData.activeColor == 'w' and literals.ltWhiteToMove or literals.ltBlackToMove
      )
    end
  end
end

function love.update(dt)
  if gameOver then
    return
  end
  
  local mx, my = love.mouse.getPosition()
  
  mx = mx + 16
  my = my + 16
  
  if draggable == 0 then
    if love.mouse.isDown(1) then
      for k, v in ipairs(pieces) do
        if inRange(mx, v.x, v.x + 40 - 1) and inRange(my, v.y, v.y + 40 - 1) then
          draggable = k
          offset.x, offset.y = mx - v.x, my - v.y
          break
        end
      end
      if draggable ~= 0 then
        move = squareToStr(screenToBoard(mx, my))
        love.mouse.setCursor(cursorGrabbing)
        erase(validTargetSquares)
        for k, v in ipairs(legalMoves) do
          if string.sub(v, 1, 2) == move then
            local x, y = boardToScreen(strToSquare(string.sub(v, 3, 4)))
            validTargetSquares[#validTargetSquares + 1] = {x = x, y = y}
          end
        end
      end
    end
  end
  
  if draggable ~= 0 then
    if not love.mouse.isDown(1) then
      pieces[draggable].x = align(pieces[draggable].x)
      pieces[draggable].y = align(pieces[draggable].y)
      
      move = move .. squareToStr(screenToBoard(pieces[draggable].x, pieces[draggable].y))
      
      if isLegalMove(move) then
        local result, square = isEnPassant(positionData, move)
        if result then
          for k, v in ipairs(pieces) do
            if squareToStr(screenToBoard(v.x, v.y)) == square then
              v.x = 320
            end
          end
        end
        
        local result, rookMove = isCastling(positionData, move)
        if result then
          for k, v in ipairs(pieces) do
            if squareToStr(screenToBoard(v.x, v.y)) == string.sub(rookMove, 1, 2) then
              v.x, v.y = boardToScreen(strToSquare(string.sub(rookMove, 3, 4)))
            end
          end
        end
        
        if isPromotion(positionData, move) then
          local myPromotion = 'q'
          move = move .. myPromotion
          pieces[draggable].id = string.sub(pieces[draggable].id, 1, 1) .. myPromotion
        end
        
        doMove(positionData, strToMove(move))
        print(decodePositionData(positionData) .. '\n')
        print(boardToText(positionData.piecePlacement) .. '\n')
        afterMoveProcedure()
        
        for k, v in ipairs(pieces) do
          if k ~= draggable then
            if (v.x == pieces[draggable].x) and (v.y == pieces[draggable].y) then
              v.x = 320
            end
          end
        end
      else
        local x, y = strToSquare(string.sub(move, 1, 2))
        pieces[draggable].x, pieces[draggable].y = boardToScreen(x, y)
        print(string.format('illegal move (%s)', move) .. '\n')
      end
      
      draggable = 0
      love.mouse.setCursor(cursorGrab)
    end
  end
  
  if draggable ~= 0 then
    pieces[draggable].x, pieces[draggable].y = mx - offset.x, my - offset.y
    if pieces[draggable].x < 0 - 2 then pieces[draggable].x = 0 - 2 end
    if pieces[draggable].y < 0 - 2 then pieces[draggable].y = 0 - 2 end
    if pieces[draggable].x > 280 + 2 then pieces[draggable].x = 280 + 2 end
    if pieces[draggable].y > 280 + 2 then pieces[draggable].y = 280 + 2 end
  end
end

function love.keypressed(key)
  if key == "escape" then
    love.event.quit()
  end
end

function love.draw()
  love.graphics.draw(images['board'], 0, 0)
  for k, v in ipairs(pieces) do
    if k ~= draggable then
      if inRange(v.x, 0, 280) then
        love.graphics.draw(images[v.id], v.x, v.y)
      end
    end
  end
  if draggable ~= 0 then
    for k, v in ipairs(validTargetSquares) do
      love.graphics.draw(images['valid'], v.x, v.y)
    end
    love.graphics.draw(images[pieces[draggable].id], pieces[draggable].x, pieces[draggable].y)
  end
  love.graphics.print({{0, 0, 0}, text}, 0, 320, 0, 1, 1, -3, -3)
end
